"""Module for storing transformation functions."""

import numpy as np
import PIL.Image

import scipy.ndimage.filters

import skimage.morphology
import skimage.exposure

from jicimagelib.io import AutoName, AutoWrite
from jicimagelib.image import Image
from jicimagelib.util.array import (
    normalise,
    reduce_stack,
    dtype_contract,
)

#############################################################################
# Function decorator for creating transforms.
#############################################################################

def transformation(func):
    """Function decorator to turn another function into a transformation."""
    def func_as_transformation(*args, **kwargs):
        
        # When using transforms that return new ndarrays we lose the
        # jicimagelib.image.Image type and the history of the image.
        # One therefore needs to:
        #  - Extract the history from the input jicimagelib.image.Image.
        #  - Apply the transformation, which may return a numpy ndarray.
        #  - Force the image to the jicimagelib.image.Image type.
        #  - Re-attach the extracted history
        if hasattr(args[0], 'history'):
            # Working on jicimagelib.Image.
            history = args[0].history
        else:
            # Working on something without a history, e.g. a ndarray stack.
            history = []
        image = func(*args, **kwargs)
        image = Image.from_array(image)
        image.history = history

        image.history.append('Applied {} transform'.format(func.__name__))
        if AutoWrite.on:
            fpath = AutoName.name(func)
            try:
                if AutoWrite.auto_safe_dtype:
                    safe_range_im = 255 * normalise(image)
                    pil_im = PIL.Image.fromarray(safe_range_im.astype(np.uint8))
                else:
                    pil_im = PIL.Image.fromarray(image)
            except TypeError:
                # Give a more meaningful error message.
                raise(TypeError(
                    "Cannot handle this data type: {}".format(image.dtype)))
            pil_im.save(fpath)
        return image
    return func_as_transformation

#############################################################################
# General purpose transforms.
#############################################################################

@transformation
def max_intensity_projection(stack):
    """Return maximum intensity projection of a stack.
    
    :param stack: 3D array from which to project third dimension 
    :returns: :class:`jicimagelib.image.Image`
    """
    return reduce_stack(stack, max)

@transformation
def min_intensity_projection(stack):
    """Return minimum intensity projection of a stack.
    
    :param stack: 3D array from which to project third dimension 
    :returns: :class:`jicimagelib.image.Image`
    """
    return reduce_stack(stack, min)

@transformation
@dtype_contract(input_dtype=np.float, output_dtype=np.float)
def smooth_gaussian(image, sigma=1):
    """Returns Gaussian smoothed image.

    :param image: numpy array or :class:`jicimagelib.image.Image`
    :param sigma: standard deviation
    :returns: :class:'jicimagelib.image.Image'
    """
    return scipy.ndimage.filters.gaussian_filter(image, sigma=sigma, mode="nearest")

@transformation
@dtype_contract(output_dtype=np.float)
def equalize_adaptive_clahe(image, ntiles=8, clip_limit=0.01):
    """Return contrast limited adaptive histogram equalized image.
    
    The return value is normalised to the range 0 to 1.

    :param image: numpy array or :class:`jicimagelib.image.Image` of dtype float
    :param ntiles: number of tile regions
    :param clip_limit: clipping limit in range 0 to 1,
                       higher values give more contrast
    """
    # Convert input for skimage.
    skimage_float_im = normalise(image)
    
    if np.all(skimage_float_im):
        raise(RuntimeError("Cannot equalise when there is no variation."))
    
    normalised = skimage.exposure.equalize_adapthist(skimage_float_im,
        ntiles_x=ntiles, ntiles_y=ntiles, clip_limit=clip_limit)

    assert np.max(normalised) == 1.0
    assert np.min(normalised) == 0.0

    return normalised

@transformation
@dtype_contract(output_dtype=np.bool)
def threshold_otsu(image, multiplier=1.0):
    """Return image thresholded using Otsu's method.
    """
    otsu_value = skimage.filters.threshold_otsu(image)
    return image > otsu_value * multiplier

@transformation
@dtype_contract(input_dtype=np.bool, output_dtype=np.bool)
def remove_small_objects(image, min_size=50):
    """Remove small objects from an boolean image.

    :param image: boolean numpy array or :class:`jicimagelib.image.Image`
    :returns: boolean :class:`jicimagelib.image.Image`
    """ 
    return skimage.morphology.remove_small_objects(image, min_size=min_size)



import scipy.ndimage as nd
import numpy as np
import skimage
import skimage.morphology

class Region(object):
    """Class representing a particular point of interest in an image, 
    represented as a bitmask with 1 indicating areas in the region."""

    def __init__(self, bitmap):
        bitmap_values = set(np.unique(bitmap))
        if len(bitmap_values - set([0, 1])):
            raise(ValueError('Region bitmap must have only 0 and 1 values'))

        self.bitmap = bitmap.astype(bool)

    @classmethod
    def select_from_array(cls, array, identifier):
        """Return a region from a numpy array.
        
        :param array: :class:`numpy.ndarray`
        :param identifier: value representing the region to select in the array
        :returns: :class:`jicimagelib.region.Region`
        """

        base_array = np.zeros(array.shape)
        array_coords = np.where(array == identifier)
        base_array[array_coords] = 1

        return cls(base_array)

    @property
    def inner(self):
        """Region formed by taking non-border elements."""

        inner_array = nd.morphology.binary_erosion(self.bitmap)
        return Region(inner_array)

    @property
    def border(self):
        """Region formed by taking border elements."""

        border_array = self.bitmap - self.inner.bitmap
        return Region(border_array)

    @property
    def convex_hull(self):
        """Return region representing the convex hull."""
        hull_array = skimage.morphology.convex_hull_image(self.bitmap)
        return Region(hull_array)

    @property
    def area(self):
        """Number of non-zero elements."""
        return np.count_nonzero(self.bitmap)

    @property
    def index_arrays(self):
        """All nonzero elements as a pair of arrays."""
        return np.where(self.bitmap == True)

    @property
    def points(self):
        """Return a list of points."""
        return zip(*self.index_arrays)


    @property
    def perimeter(self):
        """Return the perimiter."""
        return self.border.area

    def dilate(self, iterations=1):
        """Return a dilated region."""
        dilated_array = nd.morphology.binary_dilation(self.bitmap, 
                                                      iterations=iterations)
        return Region(dilated_array)

    def __repr__(self):
        return self.bitmap.__repr__()

    def __str__(self):
        return self.bitmap.__str__()



"""Module for geometric operations."""

import math

class Point2D(object):
    """Class representing a point in 2D space."""

    def __init__(self, a1, a2=None):
        if a2 is None:
            # We assume that we have given a sequence with x, y coordinates.
            self.x, self.y = a1
        else:
            self.x = a1
            self.y = a2

        self._set_types()

    def _set_types(self):
        """Make sure that x, y have consistent types and set dtype."""
        # If we given something that is not an int or a float we raise
        # a RuntimeError as we do not want to have to guess if the given
        # input should be interpreted as an int or a float, for example the
        # interpretation of the string "1" vs the interpretation of the string
        # "1.0".
        for c in (self.x, self.y):
            if not ( isinstance(c, int) or isinstance(c, float) ):
                raise(RuntimeError('x, y coords should be int or float'))

        if isinstance(self.x, int) and isinstance(self.y, int):
            self._dtype = "int"
        else:
            # At least one value is a float so promote both to float.
            self.x = float(self.x)
            self.y = float(self.y)
            self._dtype = "float"

    @property
    def dtype(self):
        """Return the type of the x, y coordinates as a string."""
        return self._dtype

    @property
    def magnitude(self):
        """Return the magnitude when treating the point as a vector."""
        return math.sqrt( self.x * self.x + self.y * self.y )

    @property
    def unit_vector(self):
        """Return the unit vector."""
        return Point2D( self.x / self.magnitude, self.x / self.magnitude )

    def distance(self, other):
        """Return distance to the other point."""
        tmp = self - other
        return tmp.magnitude

    def __repr__(self):
        s = "<Point2D(x={}, y={}, dtype={})>"
        if self.dtype == "float":
            s = "<Point2D(x={:.2f}, y={:.2f}, dtype={})>"
        return s.format( self.x, self.y, self.dtype)

    def __eq__(self, other):
        if self.dtype != other.dtype:
            return False 
        return self.x == other.x and self.y == other.y

    def __add__(self, other):
        return Point2D( self.x + other.x, self.y + other.y ) 

    def __sub__(self, other):
        return Point2D( self.x - other.x, self.y - other.y ) 

    def __mul__(self, other):
        return Point2D( self.x * other, self.y * other)

    def __div__(self, other):
        if isinstance(other, int):
            raise(NotImplementedError("Integer division not yet implemented"))
        return self * (1/other)

    def __len__(self):
        return 2

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        else:
            raise(IndexError())

    def __iter__(self):
        return iter( [self.x, self.y] )

    def astype(self, dtype):
        """Return a point of the specified dtype."""
        if dtype == "int":
            return Point2D( int( round(self.x, 0) ), int( round(self.y, 0) ) )
        elif dtype == "float":
            return Point2D( float(self.x), float(self.y))
        else:
            raise(RuntimeError("Invalid dtype: {}".format(dtype)))

    def astuple(self):
        """Return the x, y coordinates as a tuple."""
        return self.x, self.y



from setuptools import setup

# Importing the "multiprocessing" module is required for the "nose.collector".
# See also: http://bugs.python.org/issue15881#msg170215
try:
    import multiprocessing
except ImportError:
    pass

import jicimagelib

# Define the test runner.
# See also:
# http://fgimian.github.io/blog/2014/04/27/running-nose-tests-with-plugins-using-the-python-setuptools-test-command/
from setuptools.command.test import test as TestCommand
class NoseTestCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        # Run nose ensuring that argv simulates running nosetests directly.
        import nose
        nose.run_exit(argv=['nosetests'])

setup(name='jicimagelib',
      packages=['jicimagelib', 'jicimagelib.util'],
      version=jicimagelib.__version__,
      description='Python package designed to make it easy to work with microscopy images.',
      author='Tjelvar Olsson',
      author_email = 'tjelvar.olsson@jic.ac.uk',
      url = 'https://github.com/JIC-CSB/jicimagelib',
      download_url = 'https://github.com/JIC-CSB/jicimagelib/tarball/{}'.format(jicimagelib.__version__),
      keywords = ['microscopy', 'image analysis'],
      cmdclass={'test': NoseTestCommand},
      install_requires=[
        'numpy',
        'scipy',
        'scikit-image',
      ]
)
